"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class AccessControlObsolete(CloudFormationLintRule):
    """Check if the S3 Bucket Access Controls are configured"""

    id = "W3045"
    shortdesc = "Controlling access to an S3 bucket should be done with bucket policies"
    description = (
        "Nearly all access control configurations can be more successfully achieved with "
        "bucket policies. Consider using bucket policies instead of access control."
    )
    source_url = "https://docs.aws.amazon.com/AmazonS3/latest/userguide/about-object-ownership.html"
    tags = ["resources", "s3"]

    def __init__(self):
        super().__init__()
        self.resource_property_types.append("AWS::S3::Bucket")

    def match_resource_properties(self, properties, _, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        for scenario in cfn.get_object_without_conditions(
            properties, ["AccessControl"]
        ):
            props = scenario.get("Object")

            access_control = props.get("AccessControl")
            if access_control:
                matches.append(
                    RuleMatch(
                        path + ["AccessControl"],
                        "Consider using AWS::S3::BucketPolicy instead of AccessControl",
                    )
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Not(CloudFormationLintRule):
    """Check if Not values are correct"""

    id = "E1023"
    shortdesc = "Validation NOT function configuration"
    description = "Making sure that NOT functions are list"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-not"
    tags = ["functions", "not"]

    def match(self, cfn):
        matches = []

        fnnots = cfn.search_deep_keys("Fn::Not")
        for fnnot in fnnots:
            if not isinstance(fnnot[-1], list):
                message = "Function Not {0} should be a list"
                matches.append(
                    RuleMatch(fnnot, message.format("/".join(map(str, fnnot[:-2]))))
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json

import regex as re

from cfnlint.helpers import REGEX_DYN_REF_SSM
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Split(CloudFormationLintRule):
    """Check if Split values are correct"""

    id = "E1018"
    shortdesc = "Split validation of parameters"
    description = "Making sure the split function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-split.html"
    tags = ["functions", "split"]
    supported_functions = [
        "Fn::Base64",
        "Fn::FindInMap",
        "Fn::GetAZs",
        "Fn::GetAtt",
        "Fn::If",
        "Fn::ImportValue",
        "Fn::Join",
        "Fn::Select",
        "Fn::Sub",
        "Ref",
        "Fn::ToJsonString",
    ]

    def _test_delimiter(self, delimiter, path):
        matches = []
        if not isinstance(delimiter, str):
            message = "Split delimiter has to be of type string for {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))
        return matches

    def _check_dyn_ref_value(self, value, path):
        """Chec item type"""
        matches = []
        if isinstance(value, str):
            if re.match(REGEX_DYN_REF_SSM, value):
                message = f'Fn::Split does not support dynamic references at {"/".join(map(str, path[:]))}'
                matches.append(RuleMatch(path[:], message))

        return matches

    def _test_string(self, s, path):
        matches = []
        matches.extend(self._check_dyn_ref_value(json.dumps(s), path[:]))
        if isinstance(s, dict):
            if len(s) == 1:
                for key, _ in s.items():
                    if key not in self.supported_functions:
                        message = "Fn::Split doesn't support the function {0} at {1}"
                        matches.append(
                            RuleMatch(
                                path + [key],
                                message.format(key, "/".join(map(str, path))),
                            )
                        )
            else:
                message = "Split list of singular function or string for {0}"
                matches.append(
                    RuleMatch(path, message.format("/".join(map(str, path))))
                )
        elif not isinstance(s, str):
            message = "Split has to be of type string or valid function for {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))
        return matches

    def _test_split(self, split_obj, path):
        matches = []
        if not isinstance(split_obj, list):
            message = "Split should be an array of 2 for {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))
            return matches

        if len(split_obj) != 2:
            message = "Split should be an array of 2 for {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))
            return matches

        split_delimiter = split_obj[0]
        split_string = split_obj[1]
        matches.extend(self._test_delimiter(split_delimiter, path[:] + [0]))
        matches.extend(self._test_string(split_string, path[:] + [1]))

        return matches

    def match(self, cfn):
        matches = []

        split_objs = cfn.search_deep_keys("Fn::Split")

        for split_obj in split_objs:
            split_value_obj = split_obj[-1]
            tree = split_obj[:-1]
            matches.extend(self._test_split(split_value_obj, tree[:]))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.languageExtensions import LanguageExtensions
from cfnlint.rules import CloudFormationLintRule


class ToJsonString(CloudFormationLintRule):
    """Check if ToJsonString values are correct"""

    id = "E1031"
    shortdesc = "ToJsonString validation of parameters"
    description = "Making sure Fn::ToJsonString is configured correctly"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference.html"
    tags = ["functions", "toJsonString"]

    def match(self, cfn):
        has_language_extensions_transform = cfn.has_language_extensions_transform()
        unsupported_pseudo_parameters = ["AWS::NotificationARNs"]

        matches = []
        intrinsic_function = "Fn::ToJsonString"
        fn_toJsonString_objects = cfn.search_deep_keys(intrinsic_function)

        for fn_toJsonString_object in fn_toJsonString_objects:
            tree = fn_toJsonString_object[:-1]
            fn_toJsonString_object_value = fn_toJsonString_object[-1]
            LanguageExtensions().validate_transform_is_declared(
                has_language_extensions_transform,
                matches,
                tree,
                intrinsic_function,
            )
            LanguageExtensions().validate_type(
                fn_toJsonString_object_value, matches, tree, intrinsic_function
            )
            LanguageExtensions().validate_pseudo_parameters(
                fn_toJsonString_object_value,
                matches,
                tree,
                unsupported_pseudo_parameters,
                intrinsic_function,
            )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import datetime
import json

import regex as re

import cfnlint.helpers
from cfnlint.helpers import RESOURCE_SPECS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class JsonSize(CloudFormationLintRule):
    """Check if JSON Object Size is within the specified length"""

    id = "E3502"
    shortdesc = "Check if a JSON Object is within size limits"
    description = "Validate properties that are JSON values so that their length is within the limits"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["resources", "limits", "json"]

    def initialize(self, cfn):
        """Initialize the rule"""
        for resource_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "ResourceTypes"
        ):
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "PropertyTypes"
        ):
            self.resource_sub_property_types.append(property_type_spec)

    def _serialize_date(self, obj):
        if isinstance(obj, datetime.date):
            return obj.isoformat()
        raise TypeError(
            f"Object of type {obj.__class__.__name__} is not JSON serializable"
        )

    def check_value(self, value, path, prop, cfn, specs):
        """Check Role.AssumeRolePolicyDocument is within limits"""
        matches = []

        # pylint: disable=too-many-return-statements
        def remove_functions(obj):
            """Replaces intrinsic functions with string"""
            if isinstance(obj, dict):
                new_obj = {}
                if len(obj) == 1:
                    for k, v in obj.items():
                        if k in cfnlint.helpers.FUNCTIONS:
                            if k == "Fn::Sub":
                                if isinstance(v, str):
                                    return re.sub(r"\${.*}", "", v)
                                if isinstance(v, list):
                                    return re.sub(r"\${.*}", "", v[0])
                            else:
                                return ""
                        else:
                            new_obj[k] = remove_functions(v)
                            return new_obj
                else:
                    for k, v in obj.items():
                        new_obj[k] = remove_functions(v)
                    return new_obj
            elif isinstance(obj, list):
                new_list = []
                for v in obj:
                    new_list.append(remove_functions(v))
                return new_list

            return obj

        scenarios = cfn.get_object_without_nested_conditions(value, path)
        json_max_size = specs.get("JsonMax")
        for scenario in scenarios:
            j = remove_functions(scenario["Object"][prop])
            if isinstance(j, str):
                try:
                    j = json.loads(j)
                except:  # pylint: disable=bare-except
                    continue
            if (
                len(json.dumps(j, separators=(",", ":"), default=self._serialize_date))
                > json_max_size
            ):
                if scenario["Scenario"]:
                    message = (
                        "{0} JSON text cannot be longer than {1} characters when {2}"
                    )
                    scenario_text = " and ".join(
                        [
                            f'when condition "{k}" is {v}'
                            for (k, v) in scenario["Scenario"].items()
                        ]
                    )
                    matches.append(
                        RuleMatch(
                            path + [prop],
                            message.format(prop, json_max_size, scenario_text),
                        )
                    )
                else:
                    message = "{0} JSON text cannot be longer than {1} characters"
                    matches.append(
                        RuleMatch(
                            path + [prop],
                            message.format(prop, json_max_size),
                        )
                    )

        return matches

    def check(self, cfn, properties, specs, path):
        """Check itself"""
        matches = []
        for p_value, p_path in properties.items_safe(path[:]):
            for prop in p_value:
                if prop in specs:
                    value = specs.get(prop).get("Value", {})
                    if value:
                        value_type = value.get("ValueType", "")
                        primitive_type = specs.get(prop).get("PrimitiveType")
                        if primitive_type == "Json":
                            matches.extend(
                                self.check_value(
                                    p_value,
                                    p_path,
                                    prop,
                                    cfn,
                                    RESOURCE_SPECS.get(cfn.regions[0])
                                    .get("ValueTypes")
                                    .get(value_type, {}),
                                )
                            )
        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("PropertyTypes")
            .get(property_type, {})
            .get("Properties", {})
        )
        matches.extend(self.check(cfn, properties, specs, path))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("ResourceTypes")
            .get(resource_type, {})
            .get("Properties", {})
        )
        matches.extend(self.check(cfn, properties, specs, path))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ImportValue(CloudFormationLintRule):
    """Check if ImportValue values are correct"""

    id = "E1016"
    shortdesc = "ImportValue validation of parameters"
    description = "Making sure the ImportValue function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-importvalue.html"
    tags = ["functions", "importvalue"]

    def match(self, cfn):
        matches = []

        iv_objs = cfn.search_deep_keys("Fn::ImportValue")

        supported_functions = [
            "Fn::Base64",
            "Fn::FindInMap",
            "Fn::If",
            "Fn::Join",
            "Fn::Select",
            "Fn::Split",
            "Fn::Sub",
            "Ref",
        ]

        unsupported_locations = ["Conditions"]

        for iv_obj in iv_objs:
            iv_value = iv_obj[-1]
            tree = iv_obj[:-1]
            if iv_obj[0] in unsupported_locations:
                message = "ImportValue cannot be used inside {0} at {1}"
                matches.append(
                    RuleMatch(
                        tree, message.format(iv_obj[0], "/".join(map(str, tree[:-1])))
                    )
                )
            if isinstance(iv_value, dict):
                if len(iv_value) == 1:
                    for key, _ in iv_value.items():
                        if key not in supported_functions:
                            message = (
                                "ImportValue should be using supported function for {0}"
                            )
                            matches.append(
                                RuleMatch(
                                    tree, message.format("/".join(map(str, tree[:-1])))
                                )
                            )
                else:
                    message = "ImportValue should have one mapping for {0}"
                    matches.append(
                        RuleMatch(tree, message.format("/".join(map(str, tree[:-1]))))
                    )
            elif not isinstance(iv_value, str):
                message = "ImportValue should have supported function or string for {0}"
                matches.append(
                    RuleMatch(tree, message.format("/".join(map(str, tree))))
                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.helpers
from cfnlint.data import AdditionalSpecs
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Exclusive(CloudFormationLintRule):
    """Check Properties Resource Configuration"""

    id = "E2520"
    shortdesc = "Check Properties that are mutually exclusive"
    description = (
        "Making sure CloudFormation properties that are exclusive are not defined"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["resources"]

    def __init__(self):
        """Init"""
        super().__init__()
        exclusivespec = cfnlint.helpers.load_resource(AdditionalSpecs, "Exclusive.json")
        self.resource_types_specs = exclusivespec["ResourceTypes"]
        self.property_types_specs = exclusivespec["PropertyTypes"]
        for resource_type_spec in self.resource_types_specs:
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in self.property_types_specs:
            self.resource_sub_property_types.append(property_type_spec)

    def check(self, properties, exclusions, path, cfn):
        """Check itself"""
        matches = []
        for p_value, p_path in properties.items_safe(path[:]):
            for k, v in exclusions.items():
                property_sets = cfn.get_object_without_conditions(p_value, [k] + v)
                for property_set in property_sets:
                    obj = property_set["Object"].clean()
                    for prop in obj:
                        if prop == k:
                            for excl_property in exclusions[prop]:
                                if obj.get(excl_property):
                                    if property_set["Scenario"] is None:
                                        message = "Property {0} should NOT exist with {1} for {2}"
                                        matches.append(
                                            RuleMatch(
                                                p_path + [prop],
                                                message.format(
                                                    excl_property,
                                                    prop,
                                                    "/".join(map(str, p_path)),
                                                ),
                                            )
                                        )
                                    else:
                                        scenario_text = " and ".join(
                                            [
                                                f'when condition "{k}" is {v}'
                                                for (k, v) in property_set[
                                                    "Scenario"
                                                ].items()
                                            ]
                                        )
                                        message = "Property {0} should NOT exist with {1} {2} for {3}"
                                        matches.append(
                                            RuleMatch(
                                                p_path + [prop],
                                                message.format(
                                                    excl_property,
                                                    prop,
                                                    scenario_text,
                                                    "/".join(map(str, p_path)),
                                                ),
                                            )
                                        )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        exclusions = self.property_types_specs.get(property_type, {})
        matches.extend(self.check(properties, exclusions, path, cfn))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        exclusions = self.resource_types_specs.get(resource_type, {})
        matches.extend(self.check(properties, exclusions, path, cfn))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

import cfnlint.helpers
from cfnlint.data import AdditionalSpecs
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class BasedOnValue(CloudFormationLintRule):
    """Generic rule for checking properties based on value"""

    spec_type = ""
    message = ""

    def __init__(self):
        super().__init__()
        basedonvalue = cfnlint.helpers.load_resource(
            AdditionalSpecs, "BasedOnValue.json"
        )
        self.resource_types_specs = basedonvalue["ResourceTypes"]
        self.property_types_specs = basedonvalue["PropertyTypes"]
        for resource_type, resource_specs in self.resource_types_specs.items():
            for based_on_specs in resource_specs.values():
                for spec in based_on_specs:
                    if spec.get(self.spec_type):
                        self.resource_property_types.append(resource_type)
        for resource_type, resource_specs in self.property_types_specs.items():
            for based_on_specs in resource_specs.values():
                for spec in based_on_specs:
                    if spec.get(self.spec_type):
                        self.resource_sub_property_types.append(resource_type)

    def _check_value(self, value, spec, cfn):
        """Checks a value to see if it fits in the spec"""
        if isinstance(value, str):
            regex = spec.get("Regex")
            if regex:
                if re.match(spec.get("Regex"), value):
                    return True
        elif isinstance(value, dict):
            if len(value) == 1:
                for k, v in value.items():
                    ref = spec.get("Ref")
                    if ref:
                        if k == "Ref":
                            if isinstance(v, str):
                                return (
                                    cfn.template.get("Resources").get(v, {}).get("Type")
                                    in ref
                                )

                    getatt = spec.get("GetAtt")
                    if getatt:
                        if k == "Fn::GetAtt":
                            if isinstance(v, list):
                                restype = (
                                    cfn.template.get("Resources").get(v[0]).get("Type")
                                )
                                if restype in getatt:
                                    return getatt.get(restype) == v[1]

        return False

    # pylint: disable=unused-argument
    def _check_prop(self, prop, scenario):
        return False

    def _check_obj(self, properties, specs, path, cfn):
        matches = []
        for k, v in specs.items():
            for s in v:
                spec_values = s.get(self.spec_type)
                if spec_values:
                    property_set = [k] + spec_values
                    scenarios = cfn.get_object_without_conditions(
                        properties, property_names=property_set
                    )
                    for scenario in scenarios:
                        if self._check_value(scenario.get("Object").get(k), s, cfn):
                            for property_name in spec_values:
                                if self._check_prop(property_name, scenario):
                                    if scenario["Scenario"] is None:
                                        message = "When property '{0}' has its current value property '{1}' {2} at {3}"
                                        matches.append(
                                            RuleMatch(
                                                path,
                                                message.format(
                                                    k,
                                                    property_name,
                                                    self.message,
                                                    "/".join(map(str, path)),
                                                ),
                                            )
                                        )
                                    else:
                                        scenario_text = " and ".join(
                                            [
                                                f'when condition "{k}" is {v}'
                                                for (k, v) in scenario[
                                                    "Scenario"
                                                ].items()
                                            ]
                                        )
                                        message = "When property '{0}' has its current value property '{1}' {2} when {3}"
                                        matches.append(
                                            RuleMatch(
                                                path,
                                                message.format(
                                                    k,
                                                    property_name,
                                                    self.message,
                                                    scenario_text,
                                                ),
                                            )
                                        )

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        if properties is None:
            # covered under rule E3001.  If there are required properties properties is required first
            return matches

        # Need to get this spec
        specs = self.resource_types_specs.get(resource_type)

        matches.extend(self._check_obj(properties, specs, path, cfn))

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        if properties is None:
            # covered under rule E3001.  If there are required properties properties is required first
            return matches

        # Need to get this spec
        specs = self.property_types_specs.get(property_type)

        matches.extend(self._check_obj(properties, specs, path, cfn))

        return matches



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import warnings


def refactored(message):
    """Decorator for refactoring classes"""

    def cls_wrapper(cls):
        """Wrapper Class"""

        class Wrapped(cls):
            """Wrapped Class"""

            def __init__(self, *args, **kwargs):
                warnings.warn(message, FutureWarning)
                super().__init__(*args, **kwargs)

        return Wrapped

    return cls_wrapper



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.helpers import valid_snapshot_types
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class DeletionPolicy(CloudFormationLintRule):
    """Check Base Resource Configuration"""

    id = "E3035"
    shortdesc = "Check DeletionPolicy values for Resources"
    description = "Check that the DeletionPolicy values are valid"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-deletionpolicy.html"
    tags = ["resources", "deletionpolicy"]

    def check_value(self, key, path, res_type):
        """Check resource names for DeletionPolicy"""
        matches = []

        valid_values = ["Delete", "Retain", "RetainExceptOnCreate", "Snapshot"]

        supported_functions = [
            "Fn::Sub",
            "Fn::Select",
            "Fn::FindInMap",
            "Fn::If",
            "Ref",
        ]

        supported_functions_joined = ", ".join(supported_functions)

        if isinstance(key, dict):
            if len(key) == 1:
                for index_key, _ in key.items():
                    if index_key not in supported_functions:
                        message = (
                            "DeletionPolicy only supports one of the "
                            + supported_functions_joined
                            + " intrinsic functions for {0}"
                        )
                        matches.append(
                            RuleMatch(path, message.format("/".join(map(str, path))))
                        )
            else:
                message = "DeletionPolicy should have one mapping for {0}"
                matches.append(
                    RuleMatch(path, message.format("/".join(map(str, path))))
                )
        else:
            if not isinstance(key, (str)):
                message = "DeletionPolicy values should be of string at {0}"
                matches.append(
                    RuleMatch(path, message.format("/".join(map(str, path))))
                )
                return matches
            if key not in valid_values:
                message = "DeletionPolicy should be only one of {0} at {1}"
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            ", ".join(map(str, valid_values)), "/".join(map(str, path))
                        ),
                    )
                )
            if key == "Snapshot" and res_type not in valid_snapshot_types:
                message = (
                    "DeletionPolicy cannot be Snapshot for resources of type {0} at {1}"
                )
                matches.append(
                    RuleMatch(path, message.format(res_type, "/".join(map(str, path))))
                )

        return matches

    def match(self, cfn):
        matches = []

        resources = cfn.get_resources()

        for resource_name, resource_values in resources.items():
            deletion_policies = resource_values.get("DeletionPolicy")
            if deletion_policies:
                path = ["Resources", resource_name, "DeletionPolicy"]
                res_type = resource_values.get("Type")
                self.logger.debug(
                    "Validating DeletionPolicy for %s base configuration", resource_name
                )
                if isinstance(deletion_policies, list):
                    message = "Only one DeletionPolicy allowed per resource at {0}"
                    matches.append(
                        RuleMatch(path, message.format("/".join(map(str, path))))
                    )
                else:
                    matches.extend(self.check_value(deletion_policies, path, res_type))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Description(CloudFormationLintRule):
    """Check Template Description is only a String"""

    id = "E1004"
    shortdesc = "Template description can only be a string"
    description = "Template description can only be a string"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-description-structure.html"
    tags = ["description"]

    def match(self, cfn):
        matches = []

        description = cfn.template.get("Description")
        if not "Description" in cfn.template:
            return matches

        if not isinstance(description, str):
            message = "Description can only be a string"
            matches.append(RuleMatch(["Description"], message))
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class SnapStart(CloudFormationLintRule):
    """Check if Lambda SnapStart is properly configured"""

    id = "W2530"
    shortdesc = "Validate that SnapStart is properly configured"
    description = (
        "To properly leverage SnapStart, you must configure both the lambda function "
        "and attach a Lambda version resource"
    )
    source_url = "https://docs.aws.amazon.com/lambda/latest/dg/snapstart.html"
    tags = ["resources", "lambda"]

    def __init__(self):
        super().__init__()
        self.resource_property_types.append("AWS::Lambda::Function")

    def _check_value(self, value, path, **kwargs):
        lambda_versions = kwargs["lambda_versions"]
        cfn = kwargs["cfn"]

        if value != "PublishedVersions":
            return []

        # SnapStart is enabled, validate if version is attached
        matches = [
            v
            for v in lambda_versions
            if any(edge == path[1] for edge in cfn.graph.graph.neighbors(v))
        ]

        if len(matches) < 1:
            return [
                RuleMatch(
                    path,
                    "SnapStart is enabled but Lambda version is not attached",
                )
            ]

        return []

    def match_resource_properties(self, properties, _, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        # if there is no graph we can't validate
        if not cfn.graph:
            return matches

        lambda_versions = cfn.get_resources(["AWS::Lambda::Version"])

        for scenario in cfn.get_object_without_conditions(properties, ["SnapStart"]):
            props = scenario.get("Object")

            snap_start = props.get("SnapStart")
            if not snap_start:
                continue

            matches.extend(
                cfn.check_value(
                    snap_start,
                    "ApplyOn",
                    path,
                    check_value=self._check_value,
                    lambda_versions=lambda_versions,
                    cfn=cfn,
                )
            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Base64(CloudFormationLintRule):
    """Check if Base64 values are correct"""

    id = "E1021"
    shortdesc = "Base64 validation of parameters"
    description = "Making sure the Base64 function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-base64.html"
    tags = ["functions", "base64"]

    def match(self, cfn):
        matches = []

        base64_objs = cfn.search_deep_keys("Fn::Base64")
        for base64_obj in base64_objs:
            tree = base64_obj[:-1]
            value_obj = base64_obj[-1]
            if isinstance(value_obj, dict):
                if len(value_obj) == 1:
                    for key, _ in value_obj.items():
                        if key == "Fn::Split":
                            message = "Base64 needs a string at {0}"
                            matches.append(
                                RuleMatch(
                                    tree[:], message.format("/".join(map(str, tree)))
                                )
                            )
                else:
                    message = "Base64 needs a string not a map or list at {0}"
                    matches.append(
                        RuleMatch(tree[:], message.format("/".join(map(str, tree))))
                    )
            elif not isinstance(value_obj, str):
                message = "Base64 needs a string at {0}"
                matches.append(
                    RuleMatch(tree[:], message.format("/".join(map(str, tree))))
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class AccessControlOwnership(CloudFormationLintRule):
    """Check if the S3 Bucket Access Controls are set with Ownership rules"""

    id = "E3045"
    shortdesc = "Validate AccessControl are set with OwnershipControls"
    description = (
        "When using AccessControl other than private you must also "
        "configure OwnershipControls. The default is bucket owner "
        "enforced which disables ACLs."
    )
    source_url = "https://docs.aws.amazon.com/AmazonS3/latest/userguide/about-object-ownership.html"
    tags = ["resources", "s3"]

    def __init__(self):
        super().__init__()
        self.resource_property_types.append("AWS::S3::Bucket")
        self.valid_access_controls = [
            "Private",
            "BucketOwnerFullControl",
            "BucketOwnerRead",
        ]

    def match_resource_properties(self, properties, _, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        for scenario in cfn.get_object_without_conditions(
            properties, ["AccessControl", "OwnershipControls"]
        ):
            props = scenario.get("Object")

            access_controls = props.get("AccessControl")
            if not access_controls or access_controls in self.valid_access_controls:
                continue

            ownership_controls = props.get("OwnershipControls")
            if ownership_controls:
                rules = ownership_controls.get("Rules", [])
                if not isinstance(rules, list):
                    continue
                if len(ownership_controls.get("Rules", [])) > 0:
                    continue

            matches.append(
                RuleMatch(
                    path + ["AccessControl"],
                    "A bucket with AccessControl set should also have OwnershipControl configured",
                )
            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class SnapStartSupported(CloudFormationLintRule):
    """Check if Lambda function using SnapStart has the correct runtimes"""

    id = "E2530"
    shortdesc = "SnapStart supports the configured runtime"
    description = (
        "To properly leverage SnapStart, you must have a runtime of Java11 or greater"
    )
    source_url = "https://docs.aws.amazon.com/lambda/latest/dg/snapstart.html"
    tags = ["resources", "lambda"]

    def __init__(self):
        super().__init__()
        self.resource_property_types.append("AWS::Lambda::Function")
        self.child_rules = {"I2530": None}
        self.regions = [
            "us-east-2",
            "us-east-1",
            "us-west-1",
            "us-west-2",
            "af-south-1",
            "ap-east-1",
            "ap-southeast-3",
            "ap-south-1",
            "ap-northeast-2",
            "ap-northeast-3",
            "ap-southeast-1",
            "ap-southeast-2",
            "ap-northeast-1",
            "ca-central-1",
            "eu-central-1",
            "eu-west-1",
            "eu-west-2",
            "eu-south-1",
            "eu-west-3",
            "eu-north-1",
            "me-south-1",
            "sa-east-1",
        ]

    def match_resource_properties(self, properties, _, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        for region in cfn.regions:
            for scenario in cfn.get_object_without_conditions(
                properties, ["Runtime", "SnapStart"], region
            ):
                props = scenario.get("Object")

                runtime = props.get("Runtime")
                snap_start = props.get("SnapStart")
                if not snap_start:
                    if self.child_rules["I2530"]:
                        matches.extend(
                            self.child_rules["I2530"].validate(
                                runtime, path, region, self.regions
                            )
                        )
                    continue

                if region not in self.regions:
                    matches.append(
                        RuleMatch(
                            path + ["SnapStart"],
                            f"Region {region} does not support SnapStart enabled functions",
                        )
                    )

                if snap_start.get("ApplyOn") != "PublishedVersions":
                    continue

                # Validate runtime is a string before using startswith
                if not isinstance(runtime, str):
                    continue

                if (
                    runtime
                    and (not runtime.startswith("java"))
                    and runtime not in ["java8.al2", "java8"]
                ):
                    matches.append(
                        RuleMatch(
                            path + ["SnapStart", "ApplyOn"],
                            f"{runtime} is not supported for SnapStart enabled functions",
                        )
                    )

        return matches



from cfnlint.rules import RuleMatch


class LanguageExtensions:
    """Class for a CloudFormation languageExtensions"""

    def validate_transform_is_declared(
        self, has_language_extensions_transform, matches, tree, intrinsic_function
    ):
        if not has_language_extensions_transform:
            message = (
                "Missing Transform: Declare the AWS::LanguageExtensions Transform globally to enable use"
                " of the intrinsic function " + intrinsic_function + " at {0}"
            )
            matches.append(RuleMatch(tree[:], message.format("/".join(map(str, tree)))))
        return matches

    def validate_type(self, fn_object_val, matches, tree, intrinsic_function):
        if not isinstance(fn_object_val, dict) and not isinstance(fn_object_val, list):
            message = intrinsic_function + " needs a map or a list at {0}"
            matches.append(RuleMatch(tree[:], message.format("/".join(map(str, tree)))))
        elif len(fn_object_val) == 0:
            message = "Invalid value for " + intrinsic_function + " for {0}"
            matches.append(RuleMatch(tree[:], message.format("/".join(map(str, tree)))))
        return matches

    def _find_in_obj(self, obj):
        if isinstance(obj, dict):
            for k, v in obj.items():
                if k == "Ref":
                    yield v
                    return
                yield from self._find_in_obj(v)

        if isinstance(obj, list):
            for v in obj:
                yield from self._find_in_obj(v)

    def validate_pseudo_parameters(
        self, fn_object_val, matches, tree, pseudo_params, intrinsic_function
    ):
        if isinstance(fn_object_val, dict):
            ref = "Ref"
            ref_list = self._find_in_obj(fn_object_val)
            for ref in ref_list:
                if ref in pseudo_params:
                    message = (
                        intrinsic_function
                        + " does not support the pseudo parameter "
                        + ref
                        + " for {0}"
                    )
                    matches.append(
                        RuleMatch(tree[:], message.format("/".join(map(str, tree))))
                    )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Exists(CloudFormationLintRule):
    """Check if used Conditions are defined"""

    id = "E8002"
    shortdesc = "Check if the referenced Conditions are defined"
    description = (
        "Making sure the used conditions are actually defined in the Conditions section"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/conditions-section-structure.html"
    tags = ["conditions"]

    def match(self, cfn):
        matches = []
        ref_conditions = {}

        # Get all defined conditions
        conditions = cfn.template.get("Conditions", {})

        # Get all "If's" that reference a Condition
        iftrees = cfn.search_deep_keys("Fn::If")
        for iftree in iftrees:
            if isinstance(iftree[-1], list):
                if isinstance(iftree[-1][0], str):
                    ref_conditions[iftree[-1][0]] = iftree

        # Get resource's Conditions
        for resource_name, resource_values in cfn.get_resources().items():
            condition = resource_values.get("Condition")
            if isinstance(condition, str):  # make sure its a string
                path = ["Resources", resource_name, "Condition"]
                ref_conditions[condition] = path

        # Get conditions used by another condition
        condtrees = cfn.search_deep_keys("Condition")

        for condtree in condtrees:
            if condtree[0] == "Conditions":
                if isinstance(condtree[-1], (str)):
                    path = ["Conditions", condtree[-1]]
                    ref_conditions[condtree[-1]] = path

        # Get Output Conditions
        for _, output_values in cfn.get_outputs_valid().items():
            if "Condition" in output_values:
                path = ["Outputs", output_values["Condition"]]
                ref_conditions[output_values["Condition"]] = path

        # Check if all the conditions are defined
        for ref_condition, ref_path in ref_conditions.items():
            if ref_condition not in conditions:
                message = "Condition {0} is not defined."
                matches.append(RuleMatch(ref_path, message.format(ref_condition)))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class PropertiesTemplated(CloudFormationLintRule):
    """Check Base Resource Configuration"""

    id = "W3002"
    shortdesc = (
        "Warn when properties are configured to only work with the package command"
    )
    description = (
        "Some properties can be configured to only work with the CloudFormation"
        "package command. Warn when this is the case so user is aware."
    )
    source_url = (
        "https://docs.aws.amazon.com/cli/latest/reference/cloudformation/package.html"
    )
    tags = ["resources"]

    templated_exceptions = {
        "AWS::ApiGateway::RestApi": ["BodyS3Location"],
        "AWS::Lambda::Function": ["Code"],
        "AWS::Lambda::LayerVersion": ["Content"],
        "AWS::ElasticBeanstalk::ApplicationVersion": ["SourceBundle"],
        "AWS::StepFunctions::StateMachine": ["DefinitionS3Location"],
        "AWS::AppSync::GraphQLSchema": ["DefinitionS3Location"],
        "AWS::AppSync::Resolver": [
            "RequestMappingTemplateS3Location",
            "ResponseMappingTemplateS3Location",
        ],
        "AWS::AppSync::FunctionConfiguration": [
            "RequestMappingTemplateS3Location",
            "ResponseMappingTemplateS3Location",
        ],
        "AWS::CloudFormation::Stack": ["TemplateURL"],
        "AWS::CodeCommit::Repository": ["S3"],
    }

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types.extend(self.templated_exceptions.keys())

    def check_value(self, value, path):
        """Check the value"""
        matches = []
        if isinstance(value, str):
            if not value.startswith("s3://") and not value.startswith("https://"):
                message = f'This code may only work with `package` cli command as the property ({"/".join(map(str, path))}) is a string'
                matches.append(RuleMatch(path, message))

        return matches

    def match_resource_properties(self, properties, resourcetype, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        if cfn.has_serverless_transform():
            return []

        for key in self.templated_exceptions.get(resourcetype, []):
            matches.extend(
                cfn.check_value(
                    obj=properties, key=key, path=path[:], check_value=self.check_value
                )
            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.helpers
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Required(CloudFormationLintRule):
    """Check Required Resource Configuration"""

    id = "E3003"
    shortdesc = "Required Resource properties are missing"
    description = "Making sure that Resources properties that are required exist"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint/blob/main/docs/cfn-resource-specification.md#required"
    tags = ["resources"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resourcetypes = []
        self.propertytypes = []

    def initialize(self, cfn):
        """Initialize the rule"""
        resourcespecs = cfnlint.helpers.RESOURCE_SPECS[cfn.regions[0]]
        self.resourcetypes = resourcespecs["ResourceTypes"]
        self.propertytypes = resourcespecs["PropertyTypes"]
        self.resource_property_types = []
        self.resource_sub_property_types = []
        for resource_type_spec in self.resourcetypes.keys():
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in self.propertytypes.keys():
            self.resource_sub_property_types.append(property_type_spec)

    def _get_required_attrs_specs(self, resource_specs):
        reqr = []
        for k, v in resource_specs.get("Properties", {}).items():
            if v.get("Required", False):
                reqr.append(k)

        return reqr

    def check_obj(self, obj, required_attributes, path, _):
        matches = []

        for safe_obj, safe_path in obj.items_safe(path):
            for required_attribute in required_attributes:
                if required_attribute not in safe_obj:
                    message = "Property {0} missing at {1}"
                    matches.append(
                        RuleMatch(
                            safe_path,
                            message.format(
                                required_attribute, "/".join(map(str, safe_path))
                            ),
                        )
                    )

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        if properties is None:
            # covered under rule E3001.  If there are required properties properties is required first
            return matches

        matches.extend(
            self.check_obj(
                properties,
                self._get_required_attrs_specs(self.resourcetypes.get(resource_type)),
                path,
                cfn,
            )
        )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []
        matches.extend(
            self.check_obj(
                properties,
                self._get_required_attrs_specs(self.propertytypes.get(property_type)),
                path,
                cfn,
            )
        )

        return matches

    def match(self, cfn):
        """Check CloudFormation Properties"""
        matches = []

        for resourcename, resourcevalue in cfn.get_resources().items():
            if "Properties" in resourcevalue and "Type" in resourcevalue:
                resource_type = resourcevalue["Type"]
                if (
                    resource_type.startswith("Custom::")
                    and resource_type not in self.resourcetypes
                ):
                    resource_type = "AWS::CloudFormation::CustomResource"
                    matches.extend(
                        self.check_obj(
                            resourcevalue["Properties"],
                            self._get_required_attrs_specs(
                                self.resourcetypes.get(resource_type)
                            ),
                            ["Resources", resourcename, "Properties"],
                            cfn,
                        )
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import RuleMatch


def check_condition_list(cfn, function):
    matches = []
    # Build the list of functions
    trees = cfn.search_deep_keys(function)
    for tree in trees:
        # Test when in Conditions
        if tree[0] == "Conditions":
            value = tree[-1]
            if not isinstance(value, list):
                message = function + " must be a condition list"
                matches.append(RuleMatch(tree[:-1], message.format()))
            elif function == "Fn::Not" and not len(value) == 1:
                message = function + " must be a list of exactly 1 condition"
                matches.append(RuleMatch(tree[:-1], message.format()))
            elif function != "Fn::Not" and not 2 <= len(value) <= 10:
                message = function + " must be a list of between 2 to 10 conditions"
                matches.append(RuleMatch(tree[:-1], message.format()))
            else:
                for index, element in enumerate(value):
                    if isinstance(element, dict):
                        if len(element) == 1:
                            for element_key in element.keys():
                                if element_key not in [
                                    "Fn::And",
                                    "Fn::Or",
                                    "Fn::Not",
                                    "Condition",
                                    "Fn::Equals",
                                ]:
                                    message = (
                                        function
                                        + " list must be another valid condition"
                                    )
                                    matches.append(
                                        RuleMatch(
                                            tree[:-1] + [index, element_key],
                                            message.format(),
                                        )
                                    )
                        else:
                            message = function + " list must be another valid condition"
                            matches.append(
                                RuleMatch(tree[:-1] + [index], message.format())
                            )
                    else:
                        message = function + " list must be another valid condition"
                        matches.append(RuleMatch(tree[:-1] + [index], message.format()))
    return matches
